# Copyright 2018 Matthew Treinish
#
# This file is part of pyopnsense
#
# pyopnsense is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyopnsense is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyopnsense. If not, see <http://www.gnu.org/licenses/>.


from typing import Any, Optional, Union


class APIError(Exception):
    """Representation of the API exception."""

    def __init__(
        self,
        status_code: Optional[int] = None,
        resp_body: Optional[Union[str, dict[str, Any]]] = None,
        *args: dict[str, Any],
        **kwargs: dict[str, Any],
    ) -> None:
        """Initialize the API exception."""
        self.resp_body = resp_body
        self.status_code = status_code
        message = kwargs.get("message", resp_body)
        super(APIError, self).__init__(message, *args, **kwargs)



from typing import Optional

from pyopn.base_namespace import BaseNamespace

# Import the client class
from pyopn.core.dhcpv4 import LeasesClient, ServiceClient


class Dhcpv4Namespace(BaseNamespace):
    """Namespace for Dhcpv4-related API clients."""

    # Internal attribute for lazy initialization
    _service: Optional[ServiceClient] = None
    _leases: Optional[LeasesClient] = None

    @property
    def service(self) -> ServiceClient:
        """Access the ISC DHCPv4 service controller."""
        if not self._service:
            self._service = self._initialize_client("service", ServiceClient)
        return self._service

    @property
    def leases(self) -> LeasesClient:
        """Access the ISC DHCPv4 leases controller."""
        if not self._leases:
            self._leases = self._initialize_client("leases", LeasesClient)
        return self._leases



# Copyright 2024 Alex Christy
#
# This file is part of pyopn
#
# pyopn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyopn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyopn. If not, see <http://www.gnu.org/licenses/>.

from typing import Any

from validators import ValidationError, ip_address

from pyopn import client


class ServiceClient(client.OPNClient):
    """A client for interacting with the dhcpv4/service endpoint.

    :param str api_key: The API key to use for requests
    :param str api_secret: The API secret to use for requests
    :param str base_url: The base API endpoint for the OPNsense deployment
    """

    def reconfigure(self, data: dict[str, Any]) -> NotImplementedError:
        """Reconfigure the Dhcpv4 service."""
        msg = "This ISC DHCPv4 endpoint is non-functional see Kea DHCP."
        raise NotImplementedError(msg)
        return self._post("dhcpv4/service/reconfigure", data)

    def restart(self) -> dict[str, Any]:
        """Restart the ISC DHCPv4 service."""
        return self._post("dhcpv4/service/restart", "")

    def start(self) -> dict[str, Any]:
        """Start the ISC DHCPv4 service."""
        return self._post("dhcpv4/service/restart", "")

    def status(self) -> dict[str, Any]:
        """Return the status of the ISC Dhcvpv4 service."""
        return self._get("dhcpv4/service/status")

    def stop(self) -> dict[str, Any]:
        """Stop the ISC Dhcpv4 service."""
        return self._post("dhcpv4/service/stop", "")


class LeasesClient(client.OPNClient):
    """A client for interacting with the dhcpv4/leases endpoint.

    :param str api_key: The API key to use for requests
    :param str api_secret: The API secret to use for requests
    :param str base_url: The base API endpoint for the OPNsense deployment
    """

    def del_lease(self, ip: str) -> dict[str, Any]:
        """Delete lease in ISC DHCPv4 server."""
        if not ip_address.ipv4(ip):
            msg = f"Failed to delete DHCP IP lease. Invalid IP provided: {ip}"
            raise ValidationError(msg)
        return self._post(f"dhcpv4/leases/delLease/{ip}", "")

    def search_lease(self) -> dict[str, Any]:
        """Return all leases in the ISC DHCPv4 server."""
        return self._get("dhcpv4/leases/searchLease")



# Copyright 2018 Matthew Treinish
#
# This file is part of pyopnsense
#
# pyopnsense is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyopnsense is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyopnsense. If not, see <http://www.gnu.org/licenses/>.


import json
from unittest import mock

from pyopn import client, exceptions
from pyopn.tests import base


class TestOPNClient(base.TestCase):
    """Clas for testning the OPNClient class methods."""

    @mock.patch("requests.get")
    def test_get_success(self, request_mock: mock.MagicMock) -> None:
        """Test a successful GET request."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 200
        response_mock.text = json.dumps({"a": "body"})
        request_mock.return_value = response_mock
        opnclient = client.OPNClient("", "", "", timeout=10)
        resp = opnclient._get("fake_url")
        self.assertEqual({"a": "body"}, resp)
        request_mock.assert_called_once_with(
            "/fake_url", auth=("", ""), timeout=10, verify=False
        )

    @mock.patch("requests.get")
    def test_get_failures(self, request_mock: mock.MagicMock) -> None:
        """Test a failed GET request."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 401
        response_mock.text = json.dumps({"a": "body"})
        request_mock.return_value = response_mock
        opnclient = client.OPNClient("", "", "")
        self.assertRaises(exceptions.APIError, opnclient._get, "fake_url")
        request_mock.assert_called_once_with(
            "/fake_url", auth=("", ""), timeout=5, verify=False
        )

    @mock.patch("requests.post")
    def test_post_success(self, request_mock: mock.MagicMock) -> None:
        """Test a successful POST request with a body."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 200
        response_mock.text = json.dumps({"a": "body"})
        request_mock.return_value = response_mock
        opnclient = client.OPNClient("", "", "")
        resp = opnclient._post("fake_url", {})
        self.assertEqual({"a": "body"}, resp)
        request_mock.assert_called_once_with(
            "/fake_url", json={}, auth=("", ""), timeout=5, verify=False
        )

    @mock.patch("requests.post")
    def test_post_failures(self, request_mock: mock.MagicMock) -> None:
        """Test a failed POST request with a body."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 401
        response_mock.text = json.dumps({"a": "body"})
        request_mock.return_value = response_mock
        opnclient = client.OPNClient("", "", "")
        self.assertRaises(exceptions.APIError, opnclient._post, "fake_url", {})
        request_mock.assert_called_once_with(
            "/fake_url", json={}, auth=("", ""), timeout=5, verify=False
        )

    # Test for _post_file method
    @mock.patch("requests.post")
    def test_post_file_success(self, request_mock: mock.MagicMock) -> None:
        """Test a successful file POST request."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 200
        response_mock.text = json.dumps({"status": "success"})
        request_mock.return_value = response_mock

        # Create client instance
        opnclient = client.OPNClient("", "", "")

        # Mock open to simulate file content
        with mock.patch("builtins.open", mock.mock_open(read_data="file content")):
            resp = opnclient._post_file("fake_url", "path/to/file.csv")
            self.assertEqual({"status": "success"}, resp)

        # Check that the correct arguments were passed to requests.post
        request_mock.assert_called_once_with(
            "/fake_url",
            json={"payload": "file content", "filename": "file.csv"},
            auth=("", ""),
            timeout=5,
            verify=False,
        )

    @mock.patch("requests.post")
    def test_post_file_failure(self, request_mock: mock.MagicMock) -> None:
        """Test a failed file POST request."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 400
        response_mock.text = json.dumps({"error": "bad request"})
        request_mock.return_value = response_mock

        opnclient = client.OPNClient("", "", "")

        with mock.patch("builtins.open", mock.mock_open(read_data="file content")):
            self.assertRaises(
                exceptions.APIError,
                opnclient._post_file,
                "fake_url",
                "path/to/file.csv",
            )

        request_mock.assert_called_once_with(
            "/fake_url",
            json={"payload": "file content", "filename": "file.csv"},
            auth=("", ""),
            timeout=5,
            verify=False,
        )

    # Test for _post_csv_data method
    @mock.patch("requests.post")
    def test_post_csv_data_success(self, request_mock: mock.MagicMock) -> None:
        """Test a successful CSV data POST request."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 200
        response_mock.text = json.dumps({"status": "success"})
        request_mock.return_value = response_mock

        opnclient = client.OPNClient("", "", "")
        csv_data = "id,name\n1,John Doe\n2,Jane Doe"
        resp = opnclient._post_csv_data("fake_url", csv_data)
        self.assertEqual({"status": "success"}, resp)

        request_mock.assert_called_once_with(
            "/fake_url",
            json={"payload": csv_data, "filename": "data.csv"},
            auth=("", ""),
            timeout=5,
            verify=False,
        )

    @mock.patch("requests.post")
    def test_post_csv_data_failure(self, request_mock: mock.MagicMock) -> None:
        """Test a failed CSV data POST request."""
        response_mock = mock.MagicMock()
        response_mock.status_code = 500
        response_mock.text = json.dumps({"error": "internal server error"})
        request_mock.return_value = response_mock

        opnclient = client.OPNClient("", "", "")
        csv_data = "id,name\n1,John Doe\n2,Jane Doe"
        self.assertRaises(
            exceptions.APIError, opnclient._post_csv_data, "fake_url", csv_data
        )

        request_mock.assert_called_once_with(
            "/fake_url",
            json={"payload": csv_data, "filename": "data.csv"},
            auth=("", ""),
            timeout=5,
            verify=False,
        )



# Copyright 2018 Matthew Treinish
#
# This file is part of pyopnsense
#
# pyopnsense is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyopnsense is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyopnsense. If not, see <http://www.gnu.org/licenses/>.


import json
from typing import Any, Union

import requests
from requests.packages import urllib3

from pyopn import exceptions
from pyopn.constants import DEFAULT_TIMEOUT, HTTP_SUCCESS


class OPNClient(object):
    """Representation of the OPNsense API client."""

    def __init__(
        self,
        api_key: str,
        api_secret: str,
        base_url: str,
        verify_cert: bool = False,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        """Initialize the OPNsense API client."""
        self.api_key = api_key
        self.api_secret = api_secret
        self.base_url = base_url
        self.verify_cert = verify_cert
        if not self.verify_cert:
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        self.timeout = timeout

    def _process_response(
        self, response: requests.Response, raw: bool = False
    ) -> Union[str, dict[str, Any]]:
        """Return data from response objects.

        :param Response response: Response object to process.
        :param bool raw: If True, return the raw text response as a string.
                    If False, return the JSON response parsed as a dictionary.

        :return: A string containing the raw text response if `raw` is `True`. A dictionary of the JSON response if `raw` is `False`.
        :rtype: Union[str, dict[str, Any]]
        """
        if response.status_code in HTTP_SUCCESS:
            return response.text if raw else json.loads(response.text)
        raise exceptions.APIError(
            status_code=response.status_code, resp_body=response.text
        )

    def _get(self, endpoint: str, raw: bool = False) -> Union[str, dict[str, Any]]:
        """Send GET request to the specified endpoint.

        :param str endpoint: API endpoint to send the request to.
        :param bool raw: If True, return the raw text response as a string.
                    If False, return the JSON response parsed as a dictionary.

        :return: A string containing the raw text response if `raw` is `True`. A dictionary of the JSON response if `raw` is `False`.
        :rtype: Union[str, dict[str, Any]]
        """
        req_url = "{}/{}".format(self.base_url, endpoint)
        response = requests.get(
            req_url,
            verify=self.verify_cert,
            auth=(self.api_key, self.api_secret),
            timeout=self.timeout,
        )
        return self._process_response(response, raw)

    def _post(
        self, endpoint: str, data: dict[str, Any], raw: bool = False
    ) -> Union[dict[str, Any]]:
        """Send POST request to the specified endpoint with a JSON payload.

        :param str endpoint: API endpoint to send the request to.
        :param dict[str, Any] data: Dictionary to send as JSON body.
        :param bool raw: If True, return the raw text response as a string.
                    If False, return the JSON response parsed as a dictionary.

        :return: A string containing the raw text response if `raw` is `True`. A dictionary of the JSON response if `raw` is `False`.
        :rtype: Union[str, dict[str, Any]]
        """
        req_url = "{}/{}".format(self.base_url, endpoint)
        response = requests.post(
            req_url,
            json=data,
            verify=self.verify_cert,
            auth=(self.api_key, self.api_secret),
            timeout=self.timeout,
        )
        return self._process_response(response, raw)

    def _post_file(
        self, endpoint: str, file_path: str, raw: bool = False
    ) -> Union[str, dict[str, Any]]:
        """Upload a file to the specified endpoint as JSON payload.

        :param str endpoint: API endpoint to send the request to.
        :param str file_path: Path of file to upload.
        :param bool raw: If True, return the raw text response as a string.
                    If False, return the JSON response parsed as a dictionary.

        :return: A string containing the raw text response if `raw` is `True`. A dictionary of the JSON response if `raw` is `False`.
        :rtype: Union[str, dict[str, Any]]
        """
        req_url = f"{self.base_url}/{endpoint}"

        # Read the file content
        with open(file_path, "r") as f:
            file_content = f.read()

        # Prepare the JSON payload
        payload = {"payload": file_content, "filename": file_path.split("/")[-1]}

        response = requests.post(
            req_url,
            json=payload,  # Send as JSON
            verify=self.verify_cert,
            auth=(self.api_key, self.api_secret),
            timeout=self.timeout,
        )
        return self._process_response(response, raw)

    def _post_csv_data(
        self, endpoint: str, csv_data: str, raw: bool = False
    ) -> Union[str, dict[str, Any]]:
        """Upload CSV data to the specified endpoint as JSON payload.

        :param str endpoint: API endpoint to send the request to.
        :param str csv_data: CSV data as a string.
        :param bool raw: If True, return the raw text response as a string.
                    If False, return the JSON response parsed as a dictionary.

        :return: A string containing the raw text response if `raw` is `True`. A dictionary of the JSON response if `raw` is `False`.
        :rtype: Union[str, dict[str, Any]]
        """
        req_url = f"{self.base_url}/{endpoint}"

        # Prepare the JSON payload
        payload = {
            "payload": csv_data,
            "filename": "data.csv",  # Default filename for the uploaded data
        }

        # Send the request
        response = requests.post(
            req_url,
            json=payload,  # Send as JSON
            verify=self.verify_cert,
            auth=(self.api_key, self.api_secret),
            timeout=self.timeout,
        )
        return self._process_response(response, raw)



from typing import Optional

from pyopn.base_namespace import BaseNamespace

# Import the client class
from pyopn.core.kea import CtrlAgentClient, Dhcpv4Client, Leases4Client, ServiceClient


class KeaNamespace(BaseNamespace):
    """Namespace for Kea related API clients."""

    # Internal attribute for lazy initialization
    _ctrl_agent: Optional[CtrlAgentClient] = None
    _dhcpv4: Optional[Dhcpv4Client] = None
    _leases4: Optional[Leases4Client] = None
    _service: Optional[ServiceClient] = None

    @property
    def ctrl_agent(self) -> CtrlAgentClient:
        """Access the Kea ctrl_agent controller."""
        if not self._ctrl_agent:
            self._ctrl_agent = self._initialize_client("ctrl_agent", CtrlAgentClient)
        return self._ctrl_agent

    @property
    def dhcpv4(self) -> Dhcpv4Client:
        """Access the Kea dhcpv4 controller."""
        if not self._dhcpv4:
            self._dhcpv4 = self._initialize_client("dhcpv4", Dhcpv4Client)
        return self._dhcpv4

    @property
    def leases4(self) -> Leases4Client:
        """Access the Kea leases4 controller."""
        if not self._leases4:
            self._leases4 = self._initialize_client("leases4", Leases4Client)
        return self._leases4

    @property
    def service(self) -> ServiceClient:
        """Access the Kea service controller."""
        if not self._service:
            self._service = self._initialize_client("service", ServiceClient)
        return self._service



from typing import Any


class BaseNamespace:
    """Base class for OPNsense namespaces to manage common functionality."""

    def __init__(self, wrapper: Any) -> None:  # noqa: ANN401
        """Initialize a base namespace object."""
        self._wrapper = wrapper
        self._clients: dict[str, Any] = {}

    def _initialize_client(self, name: str, client_class: type) -> Any:  # noqa: ANN401
        """Lazy initialization of clients."""
        if name not in self._clients:
            self._clients[name] = client_class(
                self._wrapper.api_key,
                self._wrapper.api_secret,
                self._wrapper.base_url,
                self._wrapper.verify_cert,
                self._wrapper.timeout,
            )
        return self._clients[name]



# Copyright 2024 Alex Christy
#
# This file is part of pyopn
#
# pyopn is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyopn is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyopn. If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Optional

from pyopn import client


class CtrlAgentClient(client.OPNClient):
    """A client for interacting with the kea/ctrl_agent endpoints.

    :param str api_key: The API key to use for requests
    :param str api_secret: The API secret to use for requests
    :param str base_url: The base API endpoint for the OPNsense deployment
    """

    def get(self) -> dict[str, Any]:
        """Get the Kea controller agent configuration.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get("kea/ctrl_agent/get")

    def set(self, data: dict[str, Any]) -> dict[str, Any]:
        """Configure the Kea control agent.

        **Note:** Make sure to POST to the `kea/service/reconfigure` endpoint after this
        to enable changes.

        This function uses the `KeaCtrlAgent.xml` data model. For details, see:
        https://github.com/opnsense/core/blob/master/src/opnsense/mvc/app/models/OPNsense/Kea/KeaCtrlAgent.xml

        :param dict data: Python dictionary to be used for the body of the request.
            The dictionary should follow the `KeaCtrlAgent.xml` data model.

        Example:
            ```python
            # Enable Control Agent with default configuration
            data = {
                "ctrlagent": {
                    "general": {
                        "enabled":"1",
                        "http_host":"127.0.0.1",
                        "http_port":"8000"
                    }
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        return self._post("kea/ctrl_agent/set", data)


class Dhcpv4Client(client.OPNClient):
    """A client for interacting with the kea/dhcpv4 endpoints.

    :param str api_key: The API key to use for requests
    :param str api_secret: The API secret to use for requests
    :param str base_url: The base API endpoint for the OPNsense deployment
    """

    def get(self) -> dict[str, Any]:
        """Get the Kea DHCPv4 server configuration.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get("kea/dhcpv4/get")

    def set(self, data: Optional[dict[str, Any]] = None) -> dict[str, Any]:
        """Set configuration for the Kea DHCPv4 server.

        :param dict data: Python dictionary to be used for the body of the request.

            **Note on applying settings:** By default the OPNsense GUI will post the below
            data when you click the 'Apply' button after setting any DHCPv4 server settings.
            However, manually testing with an empty dictionary works for applying settings.

        Example:
            ```python
            data = {
                "dhcpv4": {
                    "general": {
                    "enabled": "0",
                    "interfaces": "",
                    "valid_lifetime": "4000",
                    "fwrules": "1"
                    },
                    "ha": {
                    "enabled": "0",
                    "this_server_name": "",
                    "max_unacked_clients": "2"
                    }
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        if not data:
            data = {}
        return self._post("kea/dhcpv4/set", data)

    def add_subnet(self, data: dict[str, Any]) -> dict[str, Any]:
        r"""Add subnet to the Kea DHCPv4 server.

        **Note:** Make sure to POST to the `kea/dhcpv4/set` and then `kea/service/reconfigure` endpoints after this
        to enable changes.

        This function uses the `KeaDhcpv4.xml` data model. For details, see:
        https://github.com/opnsense/core/blob/master/src/opnsense/mvc/app/models/OPNsense/Kea/KeaDhcpv4.xml

        :param dict data: Python dictionary to be used for the body of the request.
            The dictionary should follow the `KeaDhcpv4.xml` data model.

        Example:
            ```python
            data = {
                "subnet4": {
                    "subnet": "192.168.199.0/24",
                    "description": "Full network subnet",
                    "pools": "192.168.199.10 - 192.168.199.15\\n192.168.199.64/28",
                    "option_data_autocollect": "1",
                    "option_data": {
                        "routers": "",
                        "static_routes": "",
                        "domain_name_servers": "",
                        "domain_name": "",
                        "domain_search": "",
                        "ntp_servers": "",
                        "time_servers": "",
                        "tftp_server_name": "",
                        "boot_file_name": ""
                    },
                    "next_server": ""
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        return self._post("kea/dhcpv4/addSubnet", data)

    def del_subnet(self, uuid: str) -> dict[str, Any]:
        """Delete the subnet configuration on the Kea DHCPv4 server by UUID.

        **Note:** Make sure to POST to the `kea/service/reconfigure` endpoint after this
        to enable changes.

        :param str uuid: The UUID of the subnet to delete.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._post(f"kea/dhcpv4/delSubnet/{uuid}", "")

    def get_subnet(self, uuid: str) -> dict[str, Any]:
        """Get the configuration of a subnet on the Kea DHCPv4 server.

        :param str uuid: The UUID of the subnet to get the configuration for.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get(f"kea/dhcpv4/getSubnet/{uuid}")

    def search_subnet(self) -> dict[str, Any]:
        """Get the configured subnets for the Kea DHCPv4 server.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get("kea/dhcpv4/searchSubnet")

    def set_subnet(self, uuid: str, data: dict[str, Any]) -> dict[str, Any]:
        r"""Set subnet configuration on the Kea DHCPv4 server.

        **Note:** Make sure to POST to the `kea/dhcpv4/set` and then `kea/service/reconfigure` endpoints after this
        to apply changes.

        This function uses the `KeaDhcpv4.xml` data model. For details, see:
        https://github.com/opnsense/core/blob/master/src/opnsense/mvc/app/models/OPNsense/Kea/KeaDhcpv4.xml

        :param str uuid: The UUID of the subnet to set the configuration for.
        :param dict data: Python dictionary to be used for the body of the request.
            The dictionary should follow the `KeaDhcpv4.xml` data model.

        Example:
            ```python
            data = {
                "subnet4": {
                    "subnet": "192.168.199.0/24",
                    "description": "Full network subnet",
                    "pools": "192.168.199.10 - 192.168.199.15\\n192.168.199.64/28",
                    "option_data_autocollect": "1",
                    "option_data": {
                        "routers": "",
                        "static_routes": "",
                        "domain_name_servers": "",
                        "domain_name": "",
                        "domain_search": "",
                        "ntp_servers": "",
                        "time_servers": "",
                        "tftp_server_name": "",
                        "boot_file_name": ""
                    },
                    "next_server": ""
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        return self._post(f"kea/dhcpv4/setSubnet/{uuid}", data)

    def search_reservation(self) -> dict[str, Any]:
        """Get the configured reservations on the Kea DHCPv4 server.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get("kea/dhcpv4/searchReservation")

    def add_reservation(self, data: dict[str, Any]) -> dict[str, Any]:
        """Add reservation to the Kea DHCPv4 server.

        **Note:** Make sure to POST to the `kea/dhcpv4/set` and then `kea/service/reconfigure` endpoints after this
        to enable changes.

        This function uses the `KeaDhcpv4.xml` data model. For details, see:
        https://github.com/opnsense/core/blob/master/src/opnsense/mvc/app/models/OPNsense/Kea/KeaDhcpv4.xml

        :param dict data: Python dictionary to be used for the body of the request.

        Example:
            ```python
            data = {
                "reservation": {
                    "description": "Reservation description here",
                    "hostname": "test.local",
                    "hw_address": "02:42:46:e2:c3:ac",
                    "ip_address": "192.168.199.200",
                    "subnet": "f0e59e66-194c-4a61-b6ee-e8e67c545788"
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        return self._post("kea/dhcpv4/addReservation", data)

    def del_reservation(self, uuid: str) -> dict[str, Any]:
        """Delete reservation but UUID on the Kea DHCPv4 server.

        **Note:** Make sure to POST to the `kea/service/reconfigure` endpoint after this
        to enable changes.

        :param str uuid: UUID of the DHCP reservation to delete.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._post(f"kea/dhcpv4/delReservation/{uuid}", "")

    def download_reservations(self) -> str:
        """Download CSV of reservations on the Kea DHCPv4 server.

        :return: CSV-formatted string containing all DHCPv4 reservations configured on the Kea DHCPv4 server.
        :rtype: str
        """
        return self._get("kea/dhcpv4/downloadReservations", raw=True)

    def get_reservation(self, uuid: str) -> dict[str, Any]:
        """Get configuration of the reservation on Kea DHCPv4 server by UUID.

        :param str uuid: The UUID of the DHCP reservation to get the configuration for.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get(f"kea/dhcpv4/getReservation/{uuid}")

    def set_reservation(self, uuid: str, data: dict[str, Any]) -> dict[str, Any]:
        """Set DHCP reservation configuration on the Kea DHCPv4 server.

        **Note:** Make sure to POST to the `kea/dhcpv4/set` and then `kea/service/reconfigure` endpoints after this
        to enable changes.

        This function uses the `KeaDhcpv4.xml` data model. For details, see:
        https://github.com/opnsense/core/blob/master/src/opnsense/mvc/app/models/OPNsense/Kea/KeaDhcpv4.xml

        :param str uuid: The UUID of the reservation to set the configuration for.
        :param dict data: Python dictionary to be used for the body of the request.

        Example:
            ```python
            data = {
                "reservation": {
                    "description": "Reservation description here",
                    "hostname": "test.local",
                    "hw_address": "02:42:46:e2:c3:ac",
                    "ip_address": "192.168.199.200",
                    "subnet": "f0e59e66-194c-4a61-b6ee-e8e67c545788"
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        return self._post(f"kea/dhcpv4/setReservation/{uuid}", data)

    def upload_reservations(
        self, file_path: Optional[str] = None, data: Optional[str] = None
    ) -> dict[str, Any]:
        """Upload a CSV of DHCP reservations to the Kea DHCPv4 server.

        **Note:** Make sure to POST to the `kea/dhcpv4/set` and then `kea/service/reconfigure` endpoints after this
        to enable changes.

        :param file_path: Path to the CSV file to upload. Mutually exclusive with `data`.
        :param data: Raw CSV data as a string. Mutually exclusive with `file_path`.

            Example CSV:
            ```csv
            ip_address,hw_address,hostname,description
            192.168.199.69,02:42:46:e2:c3:a6,blah.local,"VMID Blah"
            ```

        :return: API response
        :rtype: dict[str, Any]

        :raises ValueError: If both or neither of `file_path` and `data` are provided.
        """
        if file_path and data:
            msg = "Provide either `file_path` or `data`, but not both."
            raise ValueError(msg)
        if not file_path and not data:
            msg = "You must provide either `file_path` or `data`."
            raise ValueError(msg)

        if file_path:
            return self._post_file("kea/dhcpv4/uploadReservations", file_path)
        if data:
            return self._post_csv_data("kea/dhcpv4/uploadReservations", data)
        return {"error": "No file path or data provided."}

    def search_peer(self) -> dict[str, Any]:
        """Get the configured peers for the Kea DHCPv4 server.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get("kea/dhcpv4/searchPeer")

    def get_peer(self, uuid: str) -> dict[str, Any]:
        """Get the configuration for a peer on the Kea DHCPv4 server by UUID.

        :param str uuid: The UUID of the peer to get the configuration for.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get(f"kea/dhcpv4/getPeer/{uuid}")

    def del_peer(self, uuid: str) -> dict[str, Any]:
        """Delete the peer on the Kea DHCPv4 server by UUID.

        **Note:** Make sure to POST to the `kea/service/reconfigure` endpoint after this
        to enable changes.

        :param str uuid: The UUID of the peer to delete.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._post(f"kea/dhcpv4/delPeer/{uuid}", "")

    def add_peer(self, data: dict[str, Any]) -> dict[str, Any]:
        """Add peer to the Kea DHCPv4 server.

        **Note:** Make sure to POST to the `kea/dhcpv4/set` and then `kea/service/reconfigure` endpoints after this
        to enable changes.

        This function uses the `KeaDhcpv4.xml` data model. For details, see:
        https://github.com/opnsense/core/blob/master/src/opnsense/mvc/app/models/OPNsense/Kea/KeaDhcpv4.xml

        :param dict data: Python dictionary to be used for the body of the request.

        Example:
            ```python
            data = {
                "peer": {
                    "name": "Test peer",
                    "role": "primary", # primary or standby
                    "url": "http://192.168.199.252:8001/"
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        return self._post("kea/dhcpv4/addPeer", data)

    def set_peer(self, uuid: str, data: dict[str, Any]) -> dict[str, Any]:
        """Set configuration of peer on the Kea DHCPv4 server by UUID.

        **Note:** Make sure to POST to the `kea/dhcpv4/set` and then `kea/service/reconfigure` endpoints after this
        to enable changes.

        This function uses the `KeaDhcpv4.xml` data model. For details, see:
        https://github.com/opnsense/core/blob/master/src/opnsense/mvc/app/models/OPNsense/Kea/KeaDhcpv4.xml

        :param str uuid: The UUID of the peer to set the configuration for.
        :param dict data: Python dictionary to be used for the body of the request.

        Example:
            ```python
            data = {
                "peer": {
                    "name": "Test peer",
                    "role": "primary", # primary or standby
                    "url": "http://192.168.199.252:8001/"
                }
            }
            ```

        :return: API response
        :rtype: dict[str, Any]

        """
        return self._post(f"kea/dhcpv4/setPeer/{uuid}", data)


class Leases4Client(client.OPNClient):
    """A client for interacting with the kea/leases4 endpoint.

    :param str api_key: The API key to use for requests
    :param str api_secret: The API secret to use for requests
    :param str base_url: The base API endpoint for the OPNsense deployment
    """


class ServiceClient(client.OPNClient):
    """A client for interacting with the kea/service endpoint.

    :param str api_key: The API key to use for requests
    :param str api_secret: The API secret to use for requests
    :param str base_url: The base API endpoint for the OPNsense deployment
    """

    def reconfigure(self) -> dict[str, Any]:
        """Enable/Reconfigure the Kea DHCP service with the currently set configuration.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._post("kea/service/reconfigure", "")

    def status(self) -> dict[str, Any]:
        """Get the status of the Kea DHCP service.

        :return: API response
        :rtype: dict[str, Any]
        """
        return self._get("kea/service/status")



# Copyright 2018 Matthew Treinish
#
# This file is part of pyopnsense
#
# pyopnsense is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyopnsense is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyopnsense. If not, see <http://www.gnu.org/licenses/>.

import fixtures
import testtools


class TestCase(testtools.TestCase):
    """Test case setup class."""

    def setUp(self) -> None:
        """Set up a test case."""
        super(TestCase, self).setUp()
        stdout = self.useFixture(fixtures.StringStream("stdout")).stream
        self.useFixture(fixtures.MonkeyPatch("sys.stdout", stdout))
        stderr = self.useFixture(fixtures.StringStream("stderr")).stream
        self.useFixture(fixtures.MonkeyPatch("sys.stderr", stderr))
        self.useFixture(fixtures.LoggerFixture(nuke_handlers=False, level=None))



import logging
from pathlib import Path
from typing import Optional, Union
from urllib.parse import urlparse

import validators

from pyopn.constants import DEFAULT_TIMEOUT
from pyopn.core.dhcpv4_namespace import Dhcpv4Namespace
from pyopn.core.kea_namespace import KeaNamespace

# Create a module-level logger
logger = logging.getLogger(__name__)


class OPNsenseAPI(object):
    """Wrapper class to manage namespaces and API credentials."""

    def __init__(  # noqa: PLR0913
        self,
        base_url: str,
        api_key_file: Optional[Union[str, Path]] = None,
        api_key: Optional[str] = None,
        api_secret: Optional[str] = None,
        verify_cert: bool = False,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> None:
        """Initialize OPNsense API object with API key file or API credentials."""
        # Load credentials: Cred file, directly, or throw error
        if api_key_file:
            logger.info("Initializing OPNSenseAPI with API key file.")
            self.api_key, self.api_secret = self._load_keys_from_file(api_key_file)
        elif api_key and api_secret:
            logger.info("Initializing OPNSenseAPI with API key and secret.")
            self.api_key = api_key
            self.api_secret = api_secret
        else:
            logger.error(
                "Initialization failed: Neither api_key_file nor both api_key and api_secret provided."
            )
            msg = "You must provide either an api_key_file path or both api_key and api_secret for initialization."
            raise ValueError(msg)

        if validators.url(base_url):
            self.base_url = self._format_base_url(base_url)
        else:
            msg = f"Provided OPNsense base URL is not valid: {base_url}"
            raise ValueError(msg)

        self.verify_cert = verify_cert
        self.timeout = timeout

        # Lazy initialization of namespaces
        self._namespaces = {}

    def _format_base_url(self, base_url: str) -> str:
        """Ensure that the base_url is properly formatted."""
        base_url = base_url.rstrip("/")

        if not base_url.endswith("/api"):
            parsed_url = urlparse(base_url)
            base_url = f"{parsed_url.scheme}://{parsed_url.netloc}/api"

        return base_url

    def _set_credentials(self, api_key: str, api_secret: str) -> None:
        """Update API credentials and invalidate existing namespaces."""
        self.api_key = api_key
        self.api_secret = api_secret
        self._namespaces.clear()

    def _load_keys_from_file(self, file_path: Union[str, Path]) -> tuple[str, str]:
        """Load the API key and secret from a file.

        :param str | Path file_path: Path to the file containing API credentials
        :returns: A tuple containing the API key and secret
        :rtype: tuple[str, str]
        :raises FileNotFoundError: If the file does not exist
        :raises ValueError: If the file contents are invalid or malformed
        """
        path = Path(file_path)
        if not path.is_file():
            logger.error("File not found: %s", file_path)
            msg = f"The file at {file_path} does not exist."
            raise FileNotFoundError(msg)

        logger.debug("Reading API credentials from file: %s", file_path)
        keys = {}
        try:
            with path.open("r", encoding="utf-8") as f:
                for line in f:
                    if "=" in line:
                        key, value = line.strip().split("=", 1)
                        keys[key.strip()] = value.strip().strip('"')  # Remove quotes

            if "key" not in keys or "secret" not in keys:
                logger.error(
                    "Invalid file format in %s: Missing 'key' or 'secret'.", file_path
                )
                msg = "The file must contain both 'key' and 'secret' in the format key=\"value\"."
                raise ValueError(msg)

            logger.info("API credentials successfully loaded from file: %s", file_path)
            return keys["key"], keys["secret"]
        except Exception as e:
            msg = f"Error parsing the API key file: {e}"
            logger.error(msg)
            raise ValueError(msg) from e

    @property
    def dhcpv4(self) -> Dhcpv4Namespace:
        """Access the ISC DHCPv4 module."""
        if "dhcpv4" not in self._namespaces:
            self._namespaces["dhcpv4"] = Dhcpv4Namespace(self)
        return self._namespaces["dhcpv4"]

    @property
    def kea(self) -> KeaNamespace:
        """Access the Kea DHCPv4 module."""
        if "kea" not in self._namespaces:
            self._namespaces["kea"] = KeaNamespace(self)
        return self._namespaces["kea"]
